"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var semaphore_async_await_1 = __importDefault(require("semaphore-async-await"));
var ethereumjs_util_1 = require("ethereumjs-util");
var block_1 = require("@ethereumjs/block");
var ethash_1 = __importDefault(require("@ethereumjs/ethash"));
var common_1 = __importDefault(require("@ethereumjs/common"));
var dbManager_1 = require("./dbManager");
var util_1 = require("./util");
var level = require('level-mem');
/**
 * This class stores and interacts with blocks.
 */
var Blockchain = /** @class */ (function () {
    /**
     * Creates new Blockchain object
     *
     * @param opts - An object with the options that this constructor takes. See [[BlockchainOptions]].
     */
    function Blockchain(opts) {
        if (opts === void 0) { opts = {}; }
        // Throw on chain or hardfork options removed in latest major release
        // to prevent implicit chain setup on a wrong chain
        if ('chain' in opts || 'hardfork' in opts) {
            throw new Error('Chain/hardfork options are not allowed any more on initialization');
        }
        if (opts.common) {
            this._common = opts.common;
        }
        else {
            var DEFAULT_CHAIN = 'mainnet';
            var DEFAULT_HARDFORK = 'chainstart';
            this._common = new common_1.default({
                chain: DEFAULT_CHAIN,
                hardfork: DEFAULT_HARDFORK,
            });
        }
        this._validatePow = opts.validatePow !== undefined ? opts.validatePow : true;
        this._validateBlocks = opts.validateBlocks !== undefined ? opts.validateBlocks : true;
        this.db = opts.db ? opts.db : level();
        this.dbManager = new dbManager_1.DBManager(this.db, this._common);
        if (this._validatePow) {
            this.ethash = new ethash_1.default(this.db);
        }
        this._heads = {};
        this._staleHeadBlock = false;
        this._staleHeads = [];
        this._lock = new semaphore_async_await_1.default(1);
        this._initDone = false;
    }
    Object.defineProperty(Blockchain.prototype, "meta", {
        /**
         * Returns an object with metadata about the Blockchain. It's defined for backwards compatibility.
         */
        get: function () {
            return {
                rawHead: this._headHeader,
                heads: this._heads,
                genesis: this._genesis,
            };
        },
        enumerable: false,
        configurable: true
    });
    /**
     * Fetches the meta info about the blockchain from the db. Meta info contains
     * hashes of the headerchain head, blockchain head, genesis block and iterator
     * heads.
     *
     * @hidden
     */
    Blockchain.prototype._init = function () {
        return __awaiter(this, void 0, void 0, function () {
            var genesisHash, error_1, heads, error_2, hash, error_3, hash, error_4;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        _a.trys.push([0, 2, , 4]);
                        return [4 /*yield*/, this.dbManager.numberToHash(new ethereumjs_util_1.BN(0))];
                    case 1:
                        genesisHash = _a.sent();
                        return [3 /*break*/, 4];
                    case 2:
                        error_1 = _a.sent();
                        return [4 /*yield*/, this._setCanonicalGenesisBlock()];
                    case 3:
                        _a.sent();
                        genesisHash = this._genesis;
                        if (error_1.type !== 'NotFoundError') {
                            throw error_1;
                        }
                        return [3 /*break*/, 4];
                    case 4:
                        _a.trys.push([4, 6, , 7]);
                        return [4 /*yield*/, this.dbManager.getHeads()];
                    case 5:
                        heads = _a.sent();
                        this._heads = heads;
                        return [3 /*break*/, 7];
                    case 6:
                        error_2 = _a.sent();
                        this._heads = {};
                        if (error_2.type !== 'NotFoundError') {
                            throw error_2;
                        }
                        return [3 /*break*/, 7];
                    case 7:
                        _a.trys.push([7, 9, , 10]);
                        return [4 /*yield*/, this.dbManager.getHeadHeader()];
                    case 8:
                        hash = _a.sent();
                        this._headHeader = hash;
                        return [3 /*break*/, 10];
                    case 9:
                        error_3 = _a.sent();
                        this._headHeader = genesisHash;
                        if (error_3.type !== 'NotFoundError') {
                            throw error_3;
                        }
                        return [3 /*break*/, 10];
                    case 10:
                        _a.trys.push([10, 12, , 13]);
                        return [4 /*yield*/, this.dbManager.getHeadBlock()];
                    case 11:
                        hash = _a.sent();
                        this._headBlock = hash;
                        return [3 /*break*/, 13];
                    case 12:
                        error_4 = _a.sent();
                        this._headBlock = genesisHash;
                        if (error_4.type !== 'NotFoundError') {
                            throw error_4;
                        }
                        return [3 /*break*/, 13];
                    case 13:
                        this._initDone = true;
                        return [2 /*return*/];
                }
            });
        });
    };
    /**
     * Sets the default genesis block
     *
     * @hidden
     */
    Blockchain.prototype._setCanonicalGenesisBlock = function () {
        return __awaiter(this, void 0, void 0, function () {
            var common, genesis;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        common = new common_1.default({
                            chain: this._common.chainId(),
                            hardfork: 'chainstart',
                        });
                        genesis = block_1.Block.genesis({}, { common: common });
                        return [4 /*yield*/, this._putBlockOrHeader(genesis, true)];
                    case 1:
                        _a.sent();
                        return [2 /*return*/];
                }
            });
        });
    };
    /**
     * Puts the genesis block in the database
     *
     * @param genesis - The genesis block to be added
     */
    Blockchain.prototype.putGenesis = function (genesis) {
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.putBlock(genesis, true)];
                    case 1:
                        _a.sent();
                        return [2 /*return*/];
                }
            });
        });
    };
    /**
     * Returns the specified iterator head.
     *
     * @param name - Optional name of the state root head (default: 'vm')
     */
    Blockchain.prototype.getHead = function (name) {
        if (name === void 0) { name = 'vm'; }
        return __awaiter(this, void 0, void 0, function () {
            var hash;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!!this._initDone) return [3 /*break*/, 2];
                        return [4 /*yield*/, this._init()];
                    case 1:
                        _a.sent();
                        _a.label = 2;
                    case 2:
                        hash = this._heads[name] || this._headBlock;
                        if (!hash) {
                            throw new Error('No head found.');
                        }
                        return [2 /*return*/, this.getBlock(hash)];
                }
            });
        });
    };
    /**
     * Returns the latest header in the canonical chain.
     */
    Blockchain.prototype.getLatestHeader = function () {
        return __awaiter(this, void 0, void 0, function () {
            var block;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!!this._initDone) return [3 /*break*/, 2];
                        return [4 /*yield*/, this._init()];
                    case 1:
                        _a.sent();
                        _a.label = 2;
                    case 2:
                        if (!this._headHeader) {
                            throw new Error('No head header set');
                        }
                        return [4 /*yield*/, this.getBlock(this._headHeader)];
                    case 3:
                        block = _a.sent();
                        return [2 /*return*/, block.header];
                }
            });
        });
    };
    /**
     * Returns the latest full block in the canonical chain.
     */
    Blockchain.prototype.getLatestBlock = function () {
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!!this._initDone) return [3 /*break*/, 2];
                        return [4 /*yield*/, this._init()];
                    case 1:
                        _a.sent();
                        _a.label = 2;
                    case 2:
                        if (!this._headBlock) {
                            throw new Error('No head block set');
                        }
                        return [2 /*return*/, this.getBlock(this._headBlock)];
                }
            });
        });
    };
    /**
     * Adds many blocks to the blockchain.
     *
     * @param blocks - The blocks to be added to the blockchain
     */
    Blockchain.prototype.putBlocks = function (blocks) {
        return __awaiter(this, void 0, void 0, function () {
            var i;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        i = 0;
                        _a.label = 1;
                    case 1:
                        if (!(i < blocks.length)) return [3 /*break*/, 4];
                        return [4 /*yield*/, this.putBlock(blocks[i])];
                    case 2:
                        _a.sent();
                        _a.label = 3;
                    case 3:
                        i++;
                        return [3 /*break*/, 1];
                    case 4: return [2 /*return*/];
                }
            });
        });
    };
    /**
     * Adds a block to the blockchain.
     *
     * @param block - The block to be added to the blockchain
     */
    Blockchain.prototype.putBlock = function (block, isGenesis) {
        return __awaiter(this, void 0, void 0, function () {
            var error_5;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!!this._initDone) return [3 /*break*/, 2];
                        return [4 /*yield*/, this._init()];
                    case 1:
                        _a.sent();
                        _a.label = 2;
                    case 2: return [4 /*yield*/, this._lock.wait()];
                    case 3:
                        _a.sent();
                        _a.label = 4;
                    case 4:
                        _a.trys.push([4, 6, , 7]);
                        return [4 /*yield*/, this._putBlockOrHeader(block, isGenesis)];
                    case 5:
                        _a.sent();
                        this._lock.release();
                        return [3 /*break*/, 7];
                    case 6:
                        error_5 = _a.sent();
                        this._lock.release();
                        throw error_5;
                    case 7: return [2 /*return*/];
                }
            });
        });
    };
    /**
     * Adds many headers to the blockchain.
     *
     * @param headers - The headers to be added to the blockchain
     */
    Blockchain.prototype.putHeaders = function (headers) {
        return __awaiter(this, void 0, void 0, function () {
            var i;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        i = 0;
                        _a.label = 1;
                    case 1:
                        if (!(i < headers.length)) return [3 /*break*/, 4];
                        return [4 /*yield*/, this.putHeader(headers[i])];
                    case 2:
                        _a.sent();
                        _a.label = 3;
                    case 3:
                        i++;
                        return [3 /*break*/, 1];
                    case 4: return [2 /*return*/];
                }
            });
        });
    };
    /**
     * Adds a header to the blockchain.
     *
     * @param header - The header to be added to the blockchain
     */
    Blockchain.prototype.putHeader = function (header) {
        return __awaiter(this, void 0, void 0, function () {
            var error_6;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!!this._initDone) return [3 /*break*/, 2];
                        return [4 /*yield*/, this._init()];
                    case 1:
                        _a.sent();
                        _a.label = 2;
                    case 2: return [4 /*yield*/, this._lock.wait()];
                    case 3:
                        _a.sent();
                        _a.label = 4;
                    case 4:
                        _a.trys.push([4, 6, , 7]);
                        return [4 /*yield*/, this._putBlockOrHeader(header)];
                    case 5:
                        _a.sent();
                        this._lock.release();
                        return [3 /*break*/, 7];
                    case 6:
                        error_6 = _a.sent();
                        this._lock.release();
                        throw error_6;
                    case 7: return [2 /*return*/];
                }
            });
        });
    };
    /**
     * @hidden
     */
    Blockchain.prototype._putBlockOrHeader = function (item, isGenesis) {
        return __awaiter(this, void 0, void 0, function () {
            var block, hash, header, number, td, currentTd, dbOps, valid, _a, _b, parentTd, rebuildInfo, ops;
            var _this = this;
            return __generator(this, function (_c) {
                switch (_c.label) {
                    case 0:
                        block = item instanceof block_1.BlockHeader ? new block_1.Block(item) : item;
                        hash = block.hash();
                        header = block.header;
                        number = header.number;
                        td = header.difficulty.clone();
                        currentTd = { header: new ethereumjs_util_1.BN(0), block: new ethereumjs_util_1.BN(0) };
                        dbOps = [];
                        if (block._common.chainId() !== this._common.chainId()) {
                            throw new Error('Chain mismatch while trying to put block or header');
                        }
                        if (!(this._validateBlocks && !isGenesis && !block.isGenesis())) return [3 /*break*/, 2];
                        return [4 /*yield*/, block.validate(this)];
                    case 1:
                        _c.sent();
                        _c.label = 2;
                    case 2:
                        if (!(this._validatePow && this.ethash)) return [3 /*break*/, 4];
                        return [4 /*yield*/, this.ethash.verifyPOW(block)];
                    case 3:
                        valid = _c.sent();
                        if (!valid) {
                            throw new Error('invalid POW');
                        }
                        _c.label = 4;
                    case 4:
                        if (!!isGenesis) return [3 /*break*/, 10];
                        if (!this._headHeader) return [3 /*break*/, 6];
                        _a = currentTd;
                        return [4 /*yield*/, this._getTd(this._headHeader)];
                    case 5:
                        _a.header = _c.sent();
                        _c.label = 6;
                    case 6:
                        if (!this._headBlock) return [3 /*break*/, 8];
                        _b = currentTd;
                        return [4 /*yield*/, this._getTd(this._headBlock)];
                    case 7:
                        _b.block = _c.sent();
                        _c.label = 8;
                    case 8: return [4 /*yield*/, this._getTd(header.parentHash, number.subn(1))];
                    case 9:
                        parentTd = _c.sent();
                        td.iadd(parentTd);
                        _c.label = 10;
                    case 10:
                        rebuildInfo = function () { return __awaiter(_this, void 0, void 0, function () {
                            var type, keyEncoding, valueEncoding, key, value;
                            return __generator(this, function (_a) {
                                switch (_a.label) {
                                    case 0:
                                        type = 'put';
                                        keyEncoding = 'binary';
                                        valueEncoding = 'binary';
                                        key = util_1.tdKey(number, hash);
                                        value = ethereumjs_util_1.rlp.encode(td);
                                        dbOps.push({ type: type, key: key, value: value, keyEncoding: keyEncoding, valueEncoding: valueEncoding });
                                        this.dbManager._cache.td.set(key, value);
                                        // save header
                                        key = util_1.headerKey(number, hash);
                                        value = header.serialize();
                                        dbOps.push({ type: type, key: key, value: value, keyEncoding: keyEncoding, valueEncoding: valueEncoding });
                                        this.dbManager._cache.header.set(key, value);
                                        // store body if it exists
                                        if (isGenesis || block.transactions.length || block.uncleHeaders.length) {
                                            key = util_1.bodyKey(number, hash);
                                            value = ethereumjs_util_1.rlp.encode(block.raw().slice(1));
                                            dbOps.push({ type: type, key: key, value: value, keyEncoding: keyEncoding, valueEncoding: valueEncoding });
                                            this.dbManager._cache.body.set(key, value);
                                        }
                                        if (!(block.isGenesis() || td.gt(currentTd.header))) return [3 /*break*/, 3];
                                        this._headHeader = hash;
                                        if (item instanceof block_1.Block) {
                                            this._headBlock = hash;
                                        }
                                        if (block.isGenesis()) {
                                            this._genesis = hash;
                                        }
                                        // delete higher number assignments and overwrite stale canonical chain
                                        return [4 /*yield*/, this._deleteStaleAssignments(number.addn(1), hash, dbOps)];
                                    case 1:
                                        // delete higher number assignments and overwrite stale canonical chain
                                        _a.sent();
                                        return [4 /*yield*/, this._rebuildCanonical(header, dbOps)];
                                    case 2:
                                        _a.sent();
                                        return [3 /*break*/, 4];
                                    case 3:
                                        if (td.gt(currentTd.block) && item instanceof block_1.Block) {
                                            this._headBlock = hash;
                                        }
                                        // save hash to number lookup info even if rebuild not needed
                                        key = util_1.hashToNumberKey(hash);
                                        value = util_1.bufBE8(number);
                                        dbOps.push({ type: type, key: key, keyEncoding: keyEncoding, valueEncoding: valueEncoding, value: value });
                                        this.dbManager._cache.hashToNumber.set(key, value);
                                        _a.label = 4;
                                    case 4: return [2 /*return*/];
                                }
                            });
                        }); };
                        return [4 /*yield*/, rebuildInfo()];
                    case 11:
                        _c.sent();
                        ops = dbOps.concat(this._saveHeadOps());
                        return [4 /*yield*/, this.dbManager.batch(ops)];
                    case 12:
                        _c.sent();
                        return [2 /*return*/];
                }
            });
        });
    };
    /**
     * Gets a block by its hash.
     *
     * @param blockId - The block's hash or number
     */
    Blockchain.prototype.getBlock = function (blockId) {
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!!this._initDone) return [3 /*break*/, 2];
                        return [4 /*yield*/, this._init()];
                    case 1:
                        _a.sent();
                        _a.label = 2;
                    case 2: return [2 /*return*/, this._getBlock(blockId)];
                }
            });
        });
    };
    /**
     * @hidden
     */
    Blockchain.prototype._getBlock = function (blockId) {
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                return [2 /*return*/, this.dbManager.getBlock(blockId)];
            });
        });
    };
    /**
     * Looks up many blocks relative to blockId
     *
     * @param blockId - The block's hash or number
     * @param maxBlocks - Max number of blocks to return
     * @param skip - Number of blocks to skip apart
     * @param reverse - Fetch blocks in reverse
     */
    Blockchain.prototype.getBlocks = function (blockId, maxBlocks, skip, reverse) {
        return __awaiter(this, void 0, void 0, function () {
            var blocks, i, nextBlock;
            var _this = this;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        blocks = [];
                        i = -1;
                        nextBlock = function (blockId) { return __awaiter(_this, void 0, void 0, function () {
                            var block, error_7, nextBlockNumber;
                            return __generator(this, function (_a) {
                                switch (_a.label) {
                                    case 0:
                                        _a.trys.push([0, 2, , 3]);
                                        return [4 /*yield*/, this.getBlock(blockId)];
                                    case 1:
                                        block = _a.sent();
                                        return [3 /*break*/, 3];
                                    case 2:
                                        error_7 = _a.sent();
                                        if (error_7.type !== 'NotFoundError') {
                                            throw error_7;
                                        }
                                        return [2 /*return*/];
                                    case 3:
                                        i++;
                                        nextBlockNumber = block.header.number.addn(reverse ? -1 : 1);
                                        if (!(i !== 0 && skip && i % (skip + 1) !== 0)) return [3 /*break*/, 5];
                                        return [4 /*yield*/, nextBlock(nextBlockNumber)];
                                    case 4: return [2 /*return*/, _a.sent()];
                                    case 5:
                                        blocks.push(block);
                                        if (!(blocks.length < maxBlocks)) return [3 /*break*/, 7];
                                        return [4 /*yield*/, nextBlock(nextBlockNumber)];
                                    case 6:
                                        _a.sent();
                                        _a.label = 7;
                                    case 7: return [2 /*return*/];
                                }
                            });
                        }); };
                        return [4 /*yield*/, nextBlock(blockId)];
                    case 1:
                        _a.sent();
                        return [2 /*return*/, blocks];
                }
            });
        });
    };
    /**
     * Given an ordered array, returns an array of hashes that are not in the blockchain yet.
     *
     * @param hashes - Ordered array of hashes
     */
    Blockchain.prototype.selectNeededHashes = function (hashes) {
        return __awaiter(this, void 0, void 0, function () {
            var max, mid, min, number, error_8;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        max = hashes.length - 1;
                        mid = min = 0;
                        _a.label = 1;
                    case 1:
                        if (!(max >= min)) return [3 /*break*/, 6];
                        number = void 0;
                        _a.label = 2;
                    case 2:
                        _a.trys.push([2, 4, , 5]);
                        return [4 /*yield*/, this.dbManager.hashToNumber(hashes[mid])];
                    case 3:
                        number = _a.sent();
                        return [3 /*break*/, 5];
                    case 4:
                        error_8 = _a.sent();
                        if (error_8.type !== 'NotFoundError') {
                            throw error_8;
                        }
                        return [3 /*break*/, 5];
                    case 5:
                        if (number) {
                            min = mid + 1;
                        }
                        else {
                            max = mid - 1;
                        }
                        mid = Math.floor((min + max) / 2);
                        return [3 /*break*/, 1];
                    case 6: return [2 /*return*/, hashes.slice(min)];
                }
            });
        });
    };
    /**
     * @hidden
     */
    Blockchain.prototype._saveHeadOps = function () {
        return [
            {
                type: 'put',
                key: 'heads',
                keyEncoding: 'binary',
                valueEncoding: 'json',
                value: this._heads,
            },
            {
                type: 'put',
                key: util_1.HEAD_HEADER_KEY,
                keyEncoding: 'binary',
                valueEncoding: 'binary',
                value: this._headHeader,
            },
            {
                type: 'put',
                key: util_1.HEAD_BLOCK_KEY,
                keyEncoding: 'binary',
                valueEncoding: 'binary',
                value: this._headBlock,
            },
        ];
    };
    /**
     * @hidden
     */
    Blockchain.prototype._saveHeads = function () {
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                return [2 /*return*/, this.dbManager.batch(this._saveHeadOps())];
            });
        });
    };
    /**
     * Delete canonical number assignments for specified number and above
     *
     * @hidden
     */
    Blockchain.prototype._deleteStaleAssignments = function (number, headHash, ops) {
        var _a;
        return __awaiter(this, void 0, void 0, function () {
            var hash, error_9, type, key, keyEncoding;
            var _this = this;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        _b.trys.push([0, 2, , 3]);
                        return [4 /*yield*/, this.dbManager.numberToHash(number)];
                    case 1:
                        hash = _b.sent();
                        return [3 /*break*/, 3];
                    case 2:
                        error_9 = _b.sent();
                        if (error_9.type !== 'NotFoundError') {
                            throw error_9;
                        }
                        return [2 /*return*/];
                    case 3:
                        type = 'del';
                        key = util_1.numberToHashKey(number);
                        keyEncoding = 'binary';
                        ops.push({ type: type, key: key, keyEncoding: keyEncoding });
                        this.dbManager._cache.numberToHash.del(key);
                        // reset stale iterator heads to current canonical head
                        Object.keys(this._heads).forEach(function (name) {
                            if (_this._heads[name].equals(hash)) {
                                _this._heads[name] = headHash;
                            }
                        });
                        // reset stale headBlock to current canonical
                        if ((_a = this._headBlock) === null || _a === void 0 ? void 0 : _a.equals(hash)) {
                            this._headBlock = headHash;
                        }
                        return [4 /*yield*/, this._deleteStaleAssignments(number.addn(1), headHash, ops)];
                    case 4:
                        _b.sent();
                        return [2 /*return*/];
                }
            });
        });
    };
    /**
     * Overwrites stale canonical number assignments.
     *
     * @hidden
     */
    Blockchain.prototype._rebuildCanonical = function (header, ops) {
        var _a;
        return __awaiter(this, void 0, void 0, function () {
            var hash, number, saveLookups, staleHash, error_10, parentHeader, error_11;
            var _this = this;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        hash = header.hash();
                        number = header.number;
                        saveLookups = function (hash, number) { return __awaiter(_this, void 0, void 0, function () {
                            var type, keyEncoding, valueEncoding, key, value;
                            return __generator(this, function (_a) {
                                type = 'put';
                                keyEncoding = 'binary';
                                valueEncoding = 'binary';
                                key = util_1.numberToHashKey(number);
                                value = hash;
                                ops.push({ type: type, key: key, keyEncoding: keyEncoding, valueEncoding: valueEncoding, value: value });
                                this.dbManager._cache.numberToHash.set(key, value);
                                key = util_1.hashToNumberKey(hash);
                                value = util_1.bufBE8(number);
                                ops.push({ type: type, key: key, keyEncoding: keyEncoding, valueEncoding: valueEncoding, value: value });
                                this.dbManager._cache.hashToNumber.set(key, value);
                                return [2 /*return*/];
                            });
                        }); };
                        if (!number.isZero()) return [3 /*break*/, 2];
                        return [4 /*yield*/, saveLookups(hash, number)];
                    case 1:
                        _b.sent();
                        return [2 /*return*/];
                    case 2:
                        staleHash = null;
                        _b.label = 3;
                    case 3:
                        _b.trys.push([3, 5, , 6]);
                        return [4 /*yield*/, this.dbManager.numberToHash(number)];
                    case 4:
                        staleHash = _b.sent();
                        return [3 /*break*/, 6];
                    case 5:
                        error_10 = _b.sent();
                        if (error_10.type !== 'NotFoundError') {
                            throw error_10;
                        }
                        return [3 /*break*/, 6];
                    case 6:
                        if (!(!staleHash || !hash.equals(staleHash))) return [3 /*break*/, 13];
                        return [4 /*yield*/, saveLookups(hash, number)
                            // flag stale head for reset
                        ];
                    case 7:
                        _b.sent();
                        // flag stale head for reset
                        Object.keys(this._heads).forEach(function (name) {
                            if (staleHash && _this._heads[name].equals(staleHash)) {
                                _this._staleHeads = _this._staleHeads || [];
                                _this._staleHeads.push(name);
                            }
                        });
                        // flag stale headBlock for reset
                        if (staleHash && ((_a = this._headBlock) === null || _a === void 0 ? void 0 : _a.equals(staleHash))) {
                            this._staleHeadBlock = true;
                        }
                        _b.label = 8;
                    case 8:
                        _b.trys.push([8, 11, , 12]);
                        return [4 /*yield*/, this._getHeader(header.parentHash, number.subn(1))];
                    case 9:
                        parentHeader = _b.sent();
                        return [4 /*yield*/, this._rebuildCanonical(parentHeader, ops)];
                    case 10:
                        _b.sent();
                        return [3 /*break*/, 12];
                    case 11:
                        error_11 = _b.sent();
                        this._staleHeads = [];
                        if (error_11.type !== 'NotFoundError') {
                            throw error_11;
                        }
                        return [3 /*break*/, 12];
                    case 12: return [3 /*break*/, 14];
                    case 13:
                        // set stale heads to last previously valid canonical block
                        this._staleHeads.forEach(function (name) {
                            _this._heads[name] = hash;
                        });
                        this._staleHeads = [];
                        // set stale headBlock to last previously valid canonical block
                        if (this._staleHeadBlock) {
                            this._headBlock = hash;
                            this._staleHeadBlock = false;
                        }
                        _b.label = 14;
                    case 14: return [2 /*return*/];
                }
            });
        });
    };
    /**
     * Deletes a block from the blockchain. All child blocks in the chain are deleted and any
     * encountered heads are set to the parent block.
     *
     * @param blockHash - The hash of the block to be deleted
     */
    Blockchain.prototype.delBlock = function (blockHash) {
        return __awaiter(this, void 0, void 0, function () {
            var error_12;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!!this._initDone) return [3 /*break*/, 2];
                        return [4 /*yield*/, this._init()];
                    case 1:
                        _a.sent();
                        _a.label = 2;
                    case 2: return [4 /*yield*/, this._lock.wait()];
                    case 3:
                        _a.sent();
                        _a.label = 4;
                    case 4:
                        _a.trys.push([4, 6, , 7]);
                        return [4 /*yield*/, this._delBlock(blockHash)];
                    case 5:
                        _a.sent();
                        this._lock.release();
                        return [3 /*break*/, 7];
                    case 6:
                        error_12 = _a.sent();
                        this._lock.release();
                        throw error_12;
                    case 7: return [2 /*return*/];
                }
            });
        });
    };
    /**
     * @hidden
     */
    Blockchain.prototype._delBlock = function (blockHash) {
        return __awaiter(this, void 0, void 0, function () {
            var dbOps, header, blockHeader, blockNumber, parentHash, canonicalHash, error_13, inCanonical;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        dbOps = [];
                        return [4 /*yield*/, this._getHeader(blockHash)];
                    case 1:
                        header = _a.sent();
                        blockHeader = header;
                        blockNumber = blockHeader.number;
                        parentHash = blockHeader.parentHash;
                        canonicalHash = null;
                        _a.label = 2;
                    case 2:
                        _a.trys.push([2, 4, , 5]);
                        return [4 /*yield*/, this.dbManager.numberToHash(blockNumber)];
                    case 3:
                        canonicalHash = _a.sent();
                        return [3 /*break*/, 5];
                    case 4:
                        error_13 = _a.sent();
                        if (error_13.type !== 'NotFoundError') {
                            throw error_13;
                        }
                        return [3 /*break*/, 5];
                    case 5:
                        inCanonical = !!canonicalHash && canonicalHash.equals(blockHash);
                        // delete the block, and if block is in the canonical chain, delete all
                        // children as well
                        return [4 /*yield*/, this._delChild(blockHash, blockNumber, inCanonical ? parentHash : null, dbOps)
                            // delete all number to hash mappings for deleted block number and above
                        ];
                    case 6:
                        // delete the block, and if block is in the canonical chain, delete all
                        // children as well
                        _a.sent();
                        if (!inCanonical) return [3 /*break*/, 8];
                        return [4 /*yield*/, this._deleteStaleAssignments(blockNumber, parentHash, dbOps)];
                    case 7:
                        _a.sent();
                        _a.label = 8;
                    case 8: return [4 /*yield*/, this.dbManager.batch(dbOps)];
                    case 9:
                        _a.sent();
                        return [2 /*return*/];
                }
            });
        });
    };
    /**
     * @hidden
     */
    Blockchain.prototype._delChild = function (hash, number, headHash, ops) {
        var _a, _b;
        return __awaiter(this, void 0, void 0, function () {
            var childHeader, error_14;
            return __generator(this, function (_c) {
                switch (_c.label) {
                    case 0:
                        // delete header, body, hash to number mapping and td
                        ops.push({
                            type: 'del',
                            key: util_1.headerKey(number, hash),
                            keyEncoding: 'binary',
                        });
                        this.dbManager._cache.header.del(util_1.headerKey(number, hash));
                        ops.push({
                            type: 'del',
                            key: util_1.bodyKey(number, hash),
                            keyEncoding: 'binary',
                        });
                        this.dbManager._cache.body.del(util_1.bodyKey(number, hash));
                        ops.push({
                            type: 'del',
                            key: util_1.hashToNumberKey(hash),
                            keyEncoding: 'binary',
                        });
                        this.dbManager._cache.hashToNumber.del(util_1.hashToNumberKey(hash));
                        ops.push({
                            type: 'del',
                            key: util_1.tdKey(number, hash),
                            keyEncoding: 'binary',
                        });
                        this.dbManager._cache.td.del(util_1.tdKey(number, hash));
                        if (!headHash) {
                            return [2 /*return*/];
                        }
                        if ((_a = this._headHeader) === null || _a === void 0 ? void 0 : _a.equals(hash)) {
                            this._headHeader = headHash;
                        }
                        if ((_b = this._headBlock) === null || _b === void 0 ? void 0 : _b.equals(hash)) {
                            this._headBlock = headHash;
                        }
                        _c.label = 1;
                    case 1:
                        _c.trys.push([1, 4, , 5]);
                        return [4 /*yield*/, this._getCanonicalHeader(number.addn(1))];
                    case 2:
                        childHeader = _c.sent();
                        return [4 /*yield*/, this._delChild(childHeader.hash(), childHeader.number, headHash, ops)];
                    case 3:
                        _c.sent();
                        return [3 /*break*/, 5];
                    case 4:
                        error_14 = _c.sent();
                        if (error_14.type !== 'NotFoundError') {
                            throw error_14;
                        }
                        return [3 /*break*/, 5];
                    case 5: return [2 /*return*/];
                }
            });
        });
    };
    /**
     * Iterates through blocks starting at the specified iterator head and calls the onBlock function
     * on each block. The current location of an iterator head can be retrieved using the `getHead()`
     * method.
     *
     * @param name - Name of the state root head
     * @param onBlock - Function called on each block with params (block, reorg)
     */
    Blockchain.prototype.iterator = function (name, onBlock) {
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!!this._initDone) return [3 /*break*/, 2];
                        return [4 /*yield*/, this._init()];
                    case 1:
                        _a.sent();
                        _a.label = 2;
                    case 2: return [2 /*return*/, this._iterator(name, onBlock)];
                }
            });
        });
    };
    /**
     * @hidden
     */
    Blockchain.prototype._iterator = function (name, onBlock) {
        return __awaiter(this, void 0, void 0, function () {
            var blockHash, lastBlock, number, blockNumber, block, reorg, error_15;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        blockHash = this._heads[name] || this._genesis;
                        if (!blockHash) {
                            return [2 /*return*/];
                        }
                        return [4 /*yield*/, this.dbManager.hashToNumber(blockHash)];
                    case 1:
                        number = _a.sent();
                        blockNumber = number.addn(1);
                        _a.label = 2;
                    case 2:
                        if (!blockNumber) return [3 /*break*/, 8];
                        _a.label = 3;
                    case 3:
                        _a.trys.push([3, 6, , 7]);
                        return [4 /*yield*/, this.getBlock(blockNumber)];
                    case 4:
                        block = _a.sent();
                        if (!block) {
                            return [3 /*break*/, 8];
                        }
                        this._heads[name] = block.hash();
                        reorg = lastBlock ? lastBlock.hash().equals(block.header.parentHash) : false;
                        lastBlock = block;
                        return [4 /*yield*/, onBlock(block, reorg)];
                    case 5:
                        _a.sent();
                        blockNumber.iaddn(1);
                        return [3 /*break*/, 7];
                    case 6:
                        error_15 = _a.sent();
                        if (error_15.type !== 'NotFoundError') {
                            throw error_15;
                        }
                        return [3 /*break*/, 8];
                    case 7: return [3 /*break*/, 2];
                    case 8: return [4 /*yield*/, this._saveHeads()];
                    case 9:
                        _a.sent();
                        return [2 /*return*/];
                }
            });
        });
    };
    /**
     * Gets a header by hash and number. Header can exist outside the canonical chain
     *
     * @hidden
     */
    Blockchain.prototype._getHeader = function (hash, number) {
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!!number) return [3 /*break*/, 2];
                        return [4 /*yield*/, this.dbManager.hashToNumber(hash)];
                    case 1:
                        number = _a.sent();
                        _a.label = 2;
                    case 2: return [2 /*return*/, this.dbManager.getHeader(hash, number)];
                }
            });
        });
    };
    /**
     * Gets a header by number. Header must be in the canonical chain
     *
     * @hidden
     */
    Blockchain.prototype._getCanonicalHeader = function (number) {
        return __awaiter(this, void 0, void 0, function () {
            var hash;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.dbManager.numberToHash(number)];
                    case 1:
                        hash = _a.sent();
                        return [2 /*return*/, this._getHeader(hash, number)];
                }
            });
        });
    };
    /**
     * Gets total difficulty for a block specified by hash and number
     *
     * @hidden
     */
    Blockchain.prototype._getTd = function (hash, number) {
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!!number) return [3 /*break*/, 2];
                        return [4 /*yield*/, this.dbManager.hashToNumber(hash)];
                    case 1:
                        number = _a.sent();
                        _a.label = 2;
                    case 2: return [2 /*return*/, this.dbManager.getTd(hash, number)];
                }
            });
        });
    };
    return Blockchain;
}());
exports.default = Blockchain;
//# sourceMappingURL=index.js.map